@@ -15,6 +15,7 @@ from typing import Optional
 from typing import Tuple
 from typing import Union
 
+from chardet import detect as detect_encoding
 from filetype import guess_extension as filetype_guess_extension
 
 from .__version__ import __version__
@@ -39,7 +40,7 @@ def guess_extension(file: bytes, default: str = "") -> str:
     if not file:
         return ""
     elif (file_type := filetype_guess_extension(file)) is None:
-        return default
+        return default if default else "txt" * (detect_encoding(file[:2048]).get("encoding", None) is not None)
     elif (ext := str(file_type)) in (exts := ("zip", "octet-stream")):
         return default if default not in exts else ext
     else:
@@ -62,11 +63,11 @@ def clean_username(username: str) -> str:
 class FADatabaseTable:
     def __init__(self, database: 'FADatabase', table: str):
         self.database: 'FADatabase' = database
-        self.table: str = table
+        self.table: str = table.upper()
         self.columns_info_: List[Tuple[str, str]] = []
         self.columns_: List[str] = []
         self.column_id_: str = ""
-        self.list_columns: List[str] = list_columns.get(table, [])
+        self.list_columns: List[str] = list_columns.get(self.table, [])
 
     def __len__(self) -> int:
         return self.database.connection.execute(f"SELECT COUNT(*) FROM {self.table}").fetchone()[0]
@@ -233,9 +234,12 @@ class FADatabaseSettings(FADatabaseTable):
         self.insert({"SETTING": setting, "SVALUE": value})
 
     def read_history(self) -> List[Tuple[float, str]]:
-        return list(map(tuple, loads(self["HISTORY"])))
+        return list(map(tuple, loads(h) if (h := self["HISTORY"]) else []))
 
-    def add_history(self, command: str, time: float = datetime.now().timestamp()):
+    def add_history(self, command: str, time: float = 0):
+        if self["HISTORY"] is None:
+            return
+        time = datetime.now().timestamp() if time <= 0 else time
         self["HISTORY"] = dumps(sorted([*self.read_history(), (time, command)], key=lambda h: h[0]))
 
 
@@ -245,8 +249,8 @@ class FADatabaseSubmissions(FADatabaseTable):
             return ""
 
         ext = guess_extension(file, ext) if guess_ext else ext
-        path = join(dirname(self.database.database_path), self.database.settings["FILESFOLDER"],
-                    tiered_path(submission_id), name + f".{ext}" * bool(ext))
+        path: str = join(dirname(self.database.database_path), self.database.settings["FILESFOLDER"],
+                         tiered_path(submission_id), name + f".{ext}" * bool(ext))
 
         makedirs(dirname(path), exist_ok=True)
         open(path, "wb").write(file)
@@ -256,7 +260,7 @@ class FADatabaseSubmissions(FADatabaseTable):
     def save_submission(self, submission: Dict[str, Union[int, str, list]], file: Optional[bytes] = None):
         submission = self.format_dict(submission)
 
-        submission["FILEEXT"] = name.split(".")[-1] if "." in (name := submission["FILELINK"].split("/")[-1]) else ""
+        submission["FILEEXT"] = name.split(".")[-1] if "." in (name := submission["FILEURL"].split("/")[-1]) else ""
         submission["FILEEXT"] = self.save_submission_file(submission["ID"], file, "submission", submission["FILEEXT"])
         submission["FILESAVED"] = bool(file)
 
