@@ -101,12 +101,13 @@ def make_submissions_table(db: Connection):
         SPECIES TEXT NOT NULL,
         GENDER TEXT NOT NULL,
         RATING TEXT NOT NULL,
+        TYPE TEXT NOT NULL CHECK (TYPE IN ('image', 'music', 'text', 'flash')),
         FILEURL TEXT NOT NULL,
         FILEEXT TEXT NOT NULL,
         FILESAVED INT NOT NULL CHECK (FILESAVED in (0, 1)),
         FAVORITE TEXT NOT NULL,
         MENTIONS TEXT NOT NULL,
-        FOLDER TEXT NOT NULL CHECK (FOLDER = 'gallery' OR FOLDER = 'scraps'),
+        FOLDER TEXT NOT NULL CHECK (FOLDER IN ('gallery', 'scraps')),
         USERUPDATE INT NOT NULL CHECK (USERUPDATE in (0, 1)),
         PRIMARY KEY (ID ASC));"""
     )
