@@ -5,6 +5,7 @@ from os import makedirs
 from os.path import basename
 from os.path import dirname
 from os.path import isdir
+from os.path import isfile
 from os.path import join as path_join
 from re import Pattern
 from re import compile as re_compile
@@ -90,6 +91,11 @@ def clean_username(user: str) -> str:
     return sub(r"[^a-z0-9.~-]", "", user.lower())
 
 
+def tiered_path(id_: Union[int, str], depth: int = 5, width: int = 2) -> str:
+    id_str: str = str(int(id_)).zfill(depth * width)
+    return join(*[id_str[n:n + width] for n in range(0, depth * width, width)])
+
+
 def make_database_3(db: Connection):
     # Create submissions table
     db.execute(
@@ -717,9 +723,65 @@ def make_database_4_8(db: Connection):
     db.commit()
 
 
-def update_2_7_to_3(db: Connection) -> Connection:
-    from .database import tiered_path
+def make_database_4_9(db: Connection):
+    db.execute(
+        f"""CREATE TABLE IF NOT EXISTS USERS
+        (USERNAME TEXT UNIQUE NOT NULL CHECK (length(USERNAME) > 0),
+        FOLDERS TEXT NOT NULL,
+        PRIMARY KEY (USERNAME ASC));"""
+    )
 
+    db.execute(
+        """CREATE TABLE IF NOT EXISTS SUBMISSIONS
+        (ID INT UNIQUE NOT NULL CHECK (ID > 0),
+        AUTHOR TEXT NOT NULL CHECK (length(AUTHOR) > 0),
+        TITLE TEXT NOT NULL,
+        DATE DATE NOT NULL CHECK (DATE==strftime('%Y-%m-%d',DATE)),
+        DESCRIPTION TEXT NOT NULL,
+        TAGS TEXT NOT NULL,
+        CATEGORY TEXT NOT NULL,
+        SPECIES TEXT NOT NULL,
+        GENDER TEXT NOT NULL,
+        RATING TEXT NOT NULL,
+        TYPE TEXT NOT NULL CHECK (TYPE IN ('image', 'music', 'text', 'flash')),
+        FILEURL TEXT NOT NULL,
+        FILEEXT TEXT NOT NULL,
+        FILESAVED INT NOT NULL CHECK (FILESAVED in (0, 1)),
+        FAVORITE TEXT NOT NULL,
+        MENTIONS TEXT NOT NULL,
+        FOLDER TEXT NOT NULL CHECK (FOLDER IN ('gallery', 'scraps')),
+        USERUPDATE INT NOT NULL CHECK (USERUPDATE in (0, 1)),
+        PRIMARY KEY (ID ASC));"""
+    )
+
+    db.execute(
+        """CREATE TABLE IF NOT EXISTS JOURNALS
+        (ID INT UNIQUE NOT NULL CHECK (ID > 0),
+        AUTHOR TEXT NOT NULL CHECK (length(AUTHOR) > 0),
+        TITLE TEXT NOT NULL,
+        DATE DATE NOT NULL CHECK (DATE==strftime('%Y-%m-%d',DATE)),
+        CONTENT TEXT NOT NULL,
+        MENTIONS TEXT NOT NULL,
+        USERUPDATE INT NOT NULL CHECK (USERUPDATE in (0, 1)),
+        PRIMARY KEY (ID ASC));""",
+    )
+
+    db.execute(
+        """CREATE TABLE IF NOT EXISTS SETTINGS
+        (SETTING TEXT UNIQUE NOT NULL CHECK (length(SETTING) > 0),
+        SVALUE TEXT NOT NULL CHECK (length(SVALUE) > 0),
+        PRIMARY KEY (SETTING ASC));"""
+    )
+
+    db.execute(f"INSERT OR IGNORE INTO SETTINGS (SETTING, SVALUE) VALUES (?, ?)", ["HISTORY", "[]"])
+    db.execute(f"INSERT OR IGNORE INTO SETTINGS (SETTING, SVALUE) VALUES (?, ?)", ["COOKIES", "{}"])
+    db.execute(f"INSERT OR IGNORE INTO SETTINGS (SETTING, SVALUE) VALUES (?, ?)", ["FILESFOLDER", "FA.files"])
+    db.execute(f"INSERT OR IGNORE INTO SETTINGS (SETTING, SVALUE) VALUES (?, ?)", ["VERSION", "4.9.0"])
+
+    db.commit()
+
+
+def update_2_7_to_3(db: Connection) -> Connection:
     print("Updating 2.7.0 to 3.0.0")
     db_new: Optional[Connection] = None
 
@@ -1045,20 +1107,21 @@ def update_3_4_to_3_5(db: Connection) -> Connection:
             f"""INSERT OR REPLACE INTO db_new.SETTINGS
             SELECT * FROM SETTINGS WHERE SETTING != "HISTORY" AND SETTING != "VERSION";"""
         )
+        db.commit()
 
         # Update history
         history: List[List[str]] = json_loads(
             db.execute("SELECT SVALUE FROM SETTINGS WHERE SETTING = 'HISTORY'").fetchone()[0]
         )
         history_new: List[Tuple[float, str]] = list(map(lambda th: (float(th[0]), th[1]), history))
-        db.commit()
-        db.close()
-        db = None
-        db_new.execute("UPDATE SETTINGS SET SVALUE = ? WHERE SETTING = 'HISTORY'", json_dumps(history_new))
+        db.execute("UPDATE db_new.SETTINGS SET SVALUE = ? WHERE SETTING = 'HISTORY'", [json_dumps(history_new)])
+        db.execute("UPDATE db_new.SETTINGS SET SVALUE = ? WHERE SETTING = 'VERSION'", ["3.5.0"])
 
         # Close databases and replace old database
         print("Close databases and replace old database")
+        db.commit()
         db_new.commit()
+        db.close()
         db_new.close()
         move(db_path, path_join(dirname(db_path), "v3_4_" + basename(db_path)))
         move(db_new_path, db_path)
@@ -1164,7 +1227,7 @@ def update_4_2_to_4_3(db: Connection) -> Connection:
         db.close()
         db_new.commit()
         db_new.close()
-        move(db_path, path_join(dirname(db_path), "v4_2" + basename(db_path)))
+        move(db_path, path_join(dirname(db_path), "v4_2_" + basename(db_path)))
         move(db_new_path, db_path)
     except (BaseException, Exception) as err:
         print("Database update interrupted!")
@@ -1228,7 +1291,7 @@ def update_4_3_to_4_4(db: Connection) -> Connection:
         print("Close databases and replace old database")
         db_new.commit()
         db_new.close()
-        move(db_path, path_join(dirname(db_path), "v4_3" + basename(db_path)))
+        move(db_path, path_join(dirname(db_path), "v4_3_" + basename(db_path)))
         move(db_new_path, db_path)
     except (BaseException, Exception) as err:
         print("Database update interrupted!")
@@ -1304,21 +1367,19 @@ def update_4_4_to_4_5(db: Connection) -> Connection:
                     missing_mentions.append((user, i))
 
         if missing_mentions:
-            print("Missing submissions:", len(missing_mentions))
-            print("                     FA_v4_5_missing_mentions.txt")
+            print("Missing submissions:", len(missing_mentions), "FA_v4_5_missing_mentions.txt")
             missing_mentions.sort(key=lambda m_: (m_[0], m_[1]))
             with open(path_join(dirname(db_path), "FA_v4_5_missing_mentions.txt"), "w") as f:
                 f.write("\n".join(f"{u} {s}" for u, s in missing_mentions))
         if double_folders:
-            print("Double folders:", len(double_folders))
-            print("                FA_v4_5_double_folders.txt")
+            print("Double folders:", len(double_folders), "FA_v4_5_double_folders.txt")
             double_folders.sort(key=lambda m_: (m_[0], m_[1]))
             with open(path_join(dirname(db_path), "FA_v4_5_double_folders.txt"), "w") as f:
                 f.write("\n".join(f"{u} {s}" for u, s in double_folders))
 
         db_new.commit()
         db_new.close()
-        move(db_path, path_join(dirname(db_path), "v4_4" + basename(db_path)))
+        move(db_path, path_join(dirname(db_path), "v4_4_" + basename(db_path)))
         move(db_new_path, db_path)
     except (BaseException, Exception) as err:
         print("Database update interrupted!")
@@ -1376,7 +1437,7 @@ def update_4_5_to_4_6(db: Connection) -> Connection:
         db.close()
         db_new.commit()
         db_new.close()
-        move(db_path, path_join(dirname(db_path), "v4_5" + basename(db_path)))
+        move(db_path, path_join(dirname(db_path), "v4_5_" + basename(db_path)))
         move(db_new_path, db_path)
     except (BaseException, Exception) as err:
         print("Database update interrupted!")
@@ -1434,7 +1495,7 @@ def update_4_6_to_4_7(db: Connection) -> Connection:
 
         db_new.commit()
         db_new.close()
-        move(db_path, path_join(dirname(db_path), "v4_6" + basename(db_path)))
+        move(db_path, path_join(dirname(db_path), "v4_6_" + basename(db_path)))
         move(db_new_path, db_path)
     except (BaseException, Exception) as err:
         print("Database update interrupted!")
@@ -1476,7 +1537,7 @@ def update_4_7_to_4_8(db: Connection) -> Connection:
             SELECT * FROM JOURNALS"""
         )
         db.execute(
-            """INSERT OR IGNORE INTO db_new.SETTINGS
+            """INSERT OR REPLACE INTO db_new.SETTINGS
             SELECT * FROM SETTINGS WHERE SETTING NOT IN ('VERSION')"""
         )
 
@@ -1485,27 +1546,111 @@ def update_4_7_to_4_8(db: Connection) -> Connection:
         db = None
 
         for u, fs in db_new.execute("select USERNAME, FOLDERS from USERS"):
-            fs = "".join(sorted(f"|{f}|" for f in fs.split(",") if f))
+            fs = "".join(sorted((f"|{f}|" for f in fs.split(",") if f), key=str.lower))
             db_new.execute("update USERS set FOLDERS = ? where USERNAME = ?", (fs, u))
         db_new.commit()
         for i, ts, fs, ms in db_new.execute("select ID, TAGS, FAVORITE, MENTIONS from SUBMISSIONS"):
             if not ts and not fs and not ms:
                 continue
-            ts = "".join(sorted(f"|{t}|" for t in ts.split(",") if t))
-            fs = "".join(sorted(f"|{f}|" for f in fs.split(",") if f))
-            ms = "".join(sorted(f"|{m}|" for m in ms.split(",") if m))
+            ts = "".join(sorted((f"|{t}|" for t in ts.split(",") if t), key=str.lower))
+            fs = "".join(sorted((f"|{f}|" for f in fs.split(",") if f), key=str.lower))
+            ms = "".join(sorted((f"|{m}|" for m in ms.split(",") if m), key=str.lower))
             db_new.execute("update SUBMISSIONS set TAGS = ?, FAVORITE = ?, MENTIONS = ? where ID = ?", (ts, fs, ms, i))
         db_new.commit()
         for i, ms in db_new.execute("select ID, MENTIONS from JOURNALS where MENTIONS != ''"):
             if not ms:
                 continue
-            ms = "".join(sorted(f"|{m}|" for m in ms.split(",") if m))
+            ms = "".join(sorted((f"|{m}|" for m in ms.split(",") if m), key=str.lower))
             db_new.execute("update JOURNALS set MENTIONS = ? where ID = ?", (ms, i))
         db_new.commit()
 
         db_new.commit()
         db_new.close()
-        move(db_path, path_join(dirname(db_path), "v4_7" + basename(db_path)))
+        move(db_path, path_join(dirname(db_path), "v4_7_" + basename(db_path)))
+        move(db_new_path, db_path)
+    except (BaseException, Exception) as err:
+        print("Database update interrupted!")
+        if db is not None:
+            db.commit()
+            db.close()
+        if db_new is not None:
+            db_new.commit()
+            db_new.close()
+        raise err
+
+    return connect_database(db_path)
+
+
+def update_4_8_to_4_9(db: Connection) -> Connection:
+    print("Updating 4.8.0 to 4.9.0")
+    db_new: Optional[Connection] = None
+
+    db_path: str = dp if (dp := database_path(db)) else "FA.db"
+    db_new_path: str = path_join(dirname(db_path), "new_" + basename(db_path))
+
+    try:
+        db_new = connect_database(db_new_path)
+        make_database_4_9(db_new)
+
+        # Transfer common submissions and users data
+        print("Transfer common submissions and users data")
+        db.execute(f"ATTACH DATABASE '{db_new_path}' AS db_new")
+        db.execute(
+            """INSERT OR IGNORE INTO db_new.USERS
+            SELECT * FROM USERS"""
+        )
+        db.execute(
+            """INSERT OR IGNORE INTO db_new.SUBMISSIONS
+            SELECT ID, AUTHOR, TITLE, DATE, DESCRIPTION, TAGS, CATEGORY, 
+            SPECIES, GENDER, RATING, 'image', FILEURL, FILEEXT, FILESAVED, 
+            FAVORITE, MENTIONS, FOLDER, USERUPDATE FROM SUBMISSIONS"""
+        )
+        db.execute(
+            """INSERT OR IGNORE INTO db_new.JOURNALS
+            SELECT * FROM JOURNALS"""
+        )
+        db.execute(
+            """INSERT OR REPLACE INTO db_new.SETTINGS
+            SELECT * FROM SETTINGS WHERE SETTING NOT IN ('VERSION')"""
+        )
+
+        db.commit()
+        db.close()
+        db = None
+
+        unknown_extensions: List[Tuple[int, str]] = []
+        blank_extensions: List[Tuple[int, str]] = []
+        for i, e, s in db_new.execute("select ID, FILEEXT, FILESAVED from SUBMISSIONS"):
+            if (e := e.lower()) in ("jpg", "jpeg", "png", "gif", "tif", "tiff"):
+                pass
+            elif e in ("mp3", "wav", "mid", "midi"):
+                db_new.execute("update SUBMISSIONS set TYPE = ? where ID = ?", ("music", i))
+            elif e in ("txt", "doc", "docx", "odt", "rtf", "pdf"):
+                db_new.execute("update SUBMISSIONS set TYPE = ? where ID = ?", ("text", i))
+            elif e == "swf":
+                db_new.execute("update SUBMISSIONS set TYPE = ? where ID = ?", ("flash", i))
+            else:
+                print(f"Unknown extension: {i} '{e}'")
+                unknown_extensions.append((i, e))
+                if not e and s:
+                    print(f"Blank extension: {i} {(p := tiered_path(i))}")
+                    blank_extensions.append((i, p))
+
+        if unknown_extensions:
+            print("Unknown extensions:", len(unknown_extensions), "FA_v4_9_unknown_extensions.txt")
+            unknown_extensions.sort(key=lambda e_: e_[0])
+            with open(path_join(dirname(db_path), "FA_v4_9_unknown_extensions.txt"), "w") as f:
+                f.write("\n".join(f"{i} {e}" for i, e in unknown_extensions))
+        if blank_extensions:
+            print("Blank extensions:", len(blank_extensions), "FA_v4_9_blank_extensions.txt")
+            print("  Manually rename submission files in folders from 'submission.' to 'submission'")
+            blank_extensions.sort(key=lambda e_: e_[0])
+            with open(path_join(dirname(db_path), "FA_v4_9_blank_extensions.txt"), "w") as f:
+                f.write("\n".join(f"{i} {e}" for i, e in blank_extensions))
+
+        db_new.commit()
+        db_new.close()
+        move(db_path, path_join(dirname(db_path), "v4_8_" + basename(db_path)))
         move(db_new_path, db_path)
     except (BaseException, Exception) as err:
         print("Database update interrupted!")
@@ -1564,6 +1709,8 @@ def update_database(db: Connection, version: str) -> Connection:
         return update_database(update_4_6_to_4_7(db), version)  # 4.6.x to 4.7.0
     elif compare_versions(db_version, "4.8.0") < 0:
         return update_database(update_4_7_to_4_8(db), version)  # 4.7.x to 4.8.0
+    elif compare_versions(db_version, "4.9.0") < 0:
+        return update_database(update_4_8_to_4_9(db), version)  # 4.8.x to 4.9.0
     elif compare_versions(db_version, version) < 0:
         return update_version(db, db_version, version)  # Update to latest patch
 
